import java.io.*;
import java.util.Arrays;
import java.util.Hashtable;
//import jsc.distributions.Binomial;
////////////////////////////////////
//THIS PROGRAM TAKES A SET OF ACE FILES (FROM DIFFERENT ASSEMBLIES), PLACES THE MAPPED READS THEN TRYS TO PLACE THE UNMAPPED PAIRED READS TO GET A FULL ASSEMBLY
//THIS IS NECESSARY BECAUSE NGEN DOES NOT MAP A READ IF IT DOES NOT OVERLAP WITH THE REFERENCE, EVEN IF IT'S MATE DOES
//NOTE THAT IN THIS VERSION, THE 'BEST' REFERENCE ASSEMBLY IS DETERMINED IN AN INITIAL PREVIEWING STEP
//**THIS VERSION EXPECTS THE READS TO BE SORTED BY LOCUS (all reads mapped in any of the assemblies are extracted and put into files based on locus id, unmapped mates should also be extracted...

public class FillOutAssembly2{
  public static void main(String[] args){
      try{

	//species abbreviations assumed:
	
	//A	1	Anolis
	//H	2	Homo
	//G	3	Gallus
	//X	4	Xenopus
	//Z	5	Zebrafish
	//M	6	Mouse
	//K	7	Manakin
	//L	8	Molly
	//P	9	Pseudacris
	//S	10	Snake
	
	String specAbbrev[] = {"A","H","G","X","Z","M","K","L","P","S"};

	int speciesID=Integer.parseInt(args[0]);		//this is a number from 1 to 10 corresponding to the species abbreviations given above
	String newRefID=args[1];				//this is the one character species abbreviation (see above)
	int maxReads=0;
	int maxLocusSize=3000;

/*
	/////////////PRECOMPUTE MAXIMUM ERRORS ALLOWED USING PBINOM////////
	System.out.println("Precalculating binomial probabilities...");
	int maxErrorsAllowed[] = new int[30000];
	maxErrorsAllowed[0]=-1;
	int tempMax=1;
	for(int i=1; i<30000; i++){			//maximum of 100,000 coverage allowed
		while(1-pbinomial(tempMax,i,0.01)>0.01){	//error probability of 1% assumed, use alpha equals 0.01...
			tempMax++;
		}	
		tempMax--;	//since the last one failed, subtract it off
		if(tempMax<1){
			maxErrorsAllowed[i]=-1;		//trigger to skip site since not enough reads were sequenced to tell anything
		}else{
			maxErrorsAllowed[i]=tempMax;	
		}
	}
*/

	////////////////////this part obtains the number of reads mapped to each locus in each assembly
	
	int nAssemblies=5*4;	//5 references X 4 sequencing types: MiSeq,HiSeq500,HiSeq700,HiSeq900

	String fileNames[]=new String[nAssemblies];
	String typeNames[]=new String[nAssemblies];
	for(int i=0; i<5; i++){
		fileNames[i*4+0]=""+speciesID+"to"+specAbbrev[i]+".ace";	//MiSeq
		fileNames[i*4+1]=""+speciesID+"to"+specAbbrev[i]+"_500.ace";	//HiSeq 500
		fileNames[i*4+2]=""+speciesID+"to"+specAbbrev[i]+"_700.ace";	//HiSeq 700
		fileNames[i*4+3]=""+speciesID+"to"+specAbbrev[i]+"_900.ace";	//HiSeq 900

		typeNames[i*4+0]="Mi500";	//MiSeq
		typeNames[i*4+1]="Hi500";	//HiSeq 500
		typeNames[i*4+2]="Hi700";	//HiSeq 700
		typeNames[i*4+3]="Hi900";	//HiSeq 900
	}
	int nReadsMapped[][] = new int[nAssemblies][512];
	for(int i=0; i<nAssemblies; i++){
		for(int j=0; j<512; j++){
			nReadsMapped[i][j]=-1;	//note that reference sequence will get counted so subtract the 1 here
		}
	}

	for(int i=0; i<nAssemblies; i++){	//assume that all ace files are in the same folder
		System.out.print("\rPreviewing file "+"Assemblies/"+speciesID+"/"+fileNames[i].substring(0,fileNames[i].length()-4).split("_")[0]+"/"+fileNames[i]);

		BufferedReader brA = new BufferedReader ( new InputStreamReader(new FileInputStream(   new File("Assemblies/"+speciesID+"/"+fileNames[i].substring(0,fileNames[i].length()-4).split("_")[0]+"/"+fileNames[i]) ) ));
		String tempQ=brA.readLine();
		int locus=-1;
		while(tempQ!=null){
			if(tempQ.startsWith("CO ")){
				tempQ=tempQ.split(" ")[1];
				tempQ=tempQ.substring(1,tempQ.length()-1);
				locus=Integer.parseInt(tempQ)-1;	//make zero-based
			}else if(tempQ.startsWith("AF ")){
				nReadsMapped[i][locus]++;
			}
			tempQ=brA.readLine();
		}
		brA.close();
	}
	System.out.println();

	int bestAssembly[]=new int[512];			//index of fileName for file containing the most reads assembled for given locus
	int bestAssemblyHist[]=new int[nAssemblies];
	for(int i=0; i<nAssemblies; i++){ bestAssemblyHist[i]=0; }	
	BufferedWriter bwA = new BufferedWriter ( new OutputStreamWriter(new FileOutputStream(   new File("Assemblies/"+speciesID+"/nReadsMapped.txt") ) ));	//to store summary of nReadsMapped...
	for(int i=0; i<nAssemblies; i++){
		bwA.write(fileNames[i]+"\t");
	}
	bwA.write("\n");

	for(int j=0; j<512; j++){
		maxReads=-1;
		for(int k=0; k<nAssemblies; k++){
			if(nReadsMapped[k][j]>maxReads){
				maxReads=nReadsMapped[k][j];
				bestAssembly[j]=k;
			}
			bwA.write(nReadsMapped[k][j]+"\t");
		}
		bestAssemblyHist[bestAssembly[j]]++;
		bwA.write("\n");
	}
	bwA.flush();
	bwA.close();

	System.out.println("Assembly\tLociThisAssemblyIsBest");
	for(int j=0; j<nAssemblies; j++){
		System.out.println(fileNames[j]+"\t"+bestAssemblyHist[j]);
	}
	System.out.println();

	Locus loci[] = new Locus[512];
	for(int i=0; i<512; i++){
		loci[i]=new Locus();
	}
	Hashtable reads = new Hashtable(20000000);	//HASH TABLE USED TO STORE READ INFORMATION
	Read read,mateRead;
	String mateName="SDF";

	String tempS="";

	//go through all reference assemblies, pull out the information/sequences that mapped, fix those from the best assembly for each locus
	for(int r=0; r<nAssemblies; r++){

		String words[];
		int locus=0;

		boolean fixThisLocus=false;

		System.out.print("\rExtracting mapping information from ace file "+fileNames[r]+"...");

		BufferedReader br = new BufferedReader ( new InputStreamReader(new FileInputStream(   new File("Assemblies/"+speciesID+"/"+speciesID+"to"+(fileNames[r].split("to"))[1].charAt(0)+"/"+fileNames[r]) ) ));
		tempS=br.readLine();
		while(tempS!=null){
			if(tempS.startsWith("CO")){
				//new locus
				//CO L100A 496 473 0 U
				words=tempS.split(" ");
			
				locus=Integer.parseInt(words[1].substring(1,words[1].length()-1))-1;
				
				fixThisLocus=bestAssembly[locus]==r;

			}else if(tempS.startsWith("AF")){
				//mapped read
				//AF M25_24_A03BV_1_2_23383_18535/2 U 1
				words=tempS.split(" ");
				if(words[1].charAt(0)!='L'){		//skip if just the reference
					//first replace _ with : to make it match the read headers, also add @
					words[1]="@"+words[1].replaceAll("_",":");

					//HAS READ BEEN SEEN BEFORE?
					read=(Read)reads.get(words[1]);	//e.g. -->  M25:24:A03BV:1:2:23383:18535/2
					if(read==null){				//for the first assembly in which the read appears, create a node in the list and fill all information (final position and "fixed" will occur when reads form the best assembly are read in)
						read = new Read();
						read.locus=locus;
						read.name=words[1];
						read.runName="Mapped_"+typeNames[r];
						read.orientation=words[2].equals("U");	//need to check to see what U means!
						read.startPos=Integer.parseInt(words[3])-1+(maxLocusSize/2);	//this may change if this is not the best assembly for this locus
						read.mappedFinal=fixThisLocus;			//mapped read considered final if this is the best assembly
						loci[locus].nReads++;
						if(loci[locus].nReads==1){	//first read in locus
							loci[locus].begRead=read;
							loci[locus].endRead=read;
						}else{
							loci[locus].endRead.next=read;
							read.prev=loci[locus].endRead;
							loci[locus].endRead=read;
						}

						//this is where the forward and reverse reads get connected if they are both mapped to this locus
						if(read.name.charAt(read.name.length()-2)=='/'){	//not a merged read!
							if(read.name.charAt(read.name.length()-1)=='1'){
								mateName=read.name.substring(0,read.name.length()-1)+'2';
							}else if(read.name.charAt(read.name.length()-1)=='2'){
								mateName=read.name.substring(0,read.name.length()-1)+'1';
							}
							mateRead=(Read)reads.get(mateName);
							if(mateRead==null){
								read.mate=null;	//keep null
							}else{			//link them up!
								read.mate=mateRead;
								mateRead.mate=read;
							}
						}
						reads.put(words[1],read);
					}else if(fixThisLocus){			//this read has been observed, but for a suboptimal assembly, need to correct position
						read.mappedFinal=fixThisLocus;
						read.startPos=Integer.parseInt(words[3])-1+(maxLocusSize/2);	
					}
				}
			}else if(tempS.startsWith("RD")){
				//e.g. RD M25_24_A03BV_1_11_25345_25090/1 127 0 0
				words=tempS.split(" ");
				if(words[1].charAt(0)!='L'){			//skip the reference sequence
					words[1]="@"+words[1].replaceAll("_",":");
					read=(Read)reads.get(words[1]);
					if(read==null){
						System.out.println("SHOULD NOT BE HERE! <"+words[1]+">");
						return;
					}
					//now read in the interleaved sequence 
					tempS=br.readLine();
					if(read.seq.length()==0 || fixThisLocus){	//allow sequence to be written the first time, but rewritten only if this is the optimal assembly for this locus
						read.seq="";
						while(tempS.length()>0){
							read.seq+=tempS;
							tempS=br.readLine();
						}
					}
				}				
			}
			tempS=br.readLine();
		}
		br.close();
	}
	System.out.println();

	//at this point all reads from all assemblies should be in the list, those appearing in the best assembly should be fixed and in the correct position relative to eachother!!

	for(int i=0; i<512; i++){
		System.out.println("LOCUS "+(i+1));
		System.out.print("\tMerging assemblies...");
		//sort the reads so that the look pretty in the assembly (fixedReads grouped at top and in order by startPos) 
		sortReadsInAssembly(loci[i]);

		//first consensus using ngen mapped reads from best assembly
		createAlignment(loci[i],maxLocusSize);
		getConsensus(loci[i],maxLocusSize);			//this gets an alignment based on the reads that are in their final mapping position
	
		//checkIntegrity(loci[i]);

		//now add unmapped mates and evaluate their best match
		System.out.print("MiSeq\t");
		getUnmappedMates(loci[i],i,reads,speciesID,maxLocusSize,"s_"+speciesID+"_L"+(i+1)+"_U",typeNames[0]);		//this addes the unmapped mates to the alignment
		System.out.print("HiSeq500\t");
		getUnmappedMates(loci[i],i,reads,speciesID,maxLocusSize,"s_"+speciesID+"_500_L"+(i+1)+"_U",typeNames[1]);	//this addes the unmapped mates to the alignment
		System.out.print("HiSeq700\t");
		getUnmappedMates(loci[i],i,reads,speciesID,maxLocusSize,"s_"+speciesID+"_700_L"+(i+1)+"_U",typeNames[2]);	//this addes the unmapped mates to the alignment
		System.out.print("HiSeq900\t");
		getUnmappedMates(loci[i],i,reads,speciesID,maxLocusSize,"s_"+speciesID+"_900_L"+(i+1)+"_U",typeNames[3]);	//this addes the unmapped mates to the alignment
		System.out.println();

		System.out.print("\tRefining assemblies...");
		//checkIntegrity(loci[i]);

		getMapProbs(loci[i]);

		//checkIntegrity(loci[i]);
	
		//fix them if mapOverlap>=25 and % matched >90%
		int improvement=updateMappedFinal(loci[i]);

		//checkIntegrity(loci[i]);

		//sort the reads so that the look pretty in the assembly (fixedReads grouped at top and in order by startPos) 
		sortReadsInAssembly(loci[i]);

		//checkIntegrity(loci[i]);
			System.out.print(improvement+"\t");
			
		while(improvement>0){
			System.out.print(improvement+"\t");
			createAlignment(loci[i],maxLocusSize);
			getConsensus(loci[i],maxLocusSize);
				//checkIntegrity(loci[i]);
			refineOptimalStartLocs(loci[i],maxLocusSize);
				//checkIntegrity(loci[i]);
			getMapProbs(loci[i]);
				//checkIntegrity(loci[i]);
			improvement=updateMappedFinal(loci[i]);
				//checkIntegrity(loci[i]);
			sortReadsInAssembly(loci[i]);	
				//checkIntegrity(loci[i]);		
		}
		System.out.println();
	
		removeDuds(loci[i]);
	
		if(i==0){
			writeNReadsFixed(loci[i],speciesID,false);
		}else{
			writeNReadsFixed(loci[i],speciesID,true);
		}

		//checkIntegrity(loci[i]);	
	
		createAlignment(loci[i],maxLocusSize);	

		//System.out.println("\tCleaning up alignment...");
		//cleanUpAlignment(loci[i],maxErrorsAllowed, maxLocusSize);

		System.out.println("\tWriting final alignment...");

		//fix the alignment so that the leading and trailing Ns are '-'
		boolean foundBase;

		for(int j=0; j<loci[i].nReads; j++){
			foundBase=false;
			//check begining
			for(int k=0; k<maxLocusSize; k++){
				if(loci[i].alignment[j][k]=='N'){
					loci[i].alignment[j][k]='-';
				}else{
					foundBase=true;
					break;
				}				
			}
			
			//check end
			foundBase=false;
			//check begining
			for(int k=(maxLocusSize-1); k>=0; k--){
				if(loci[i].alignment[j][k]=='N'){
					loci[i].alignment[j][k]='-';
				}else{
					foundBase=true;
					break;
				}
			}
		}


		//write the "final" alignment to a file
		BufferedWriter bw = new BufferedWriter ( new OutputStreamWriter(new FileOutputStream(   new File("assemblies/"+speciesID+"/ReadAlignments/L"+(i+1)+newRefID+".fasta") ) ));	//output file that will contain the alignment
		read = loci[i].begRead;
		for(int j=0; j<loci[i].nReads; j++){
			if(read.mate!=null && read.mate.mappedFinal){
				bw.write(">"+read.runName+"_"+(Math.abs(read.startPos-read.mate.startPos)+225)+"_"+read.orientation+"_"+read.name);//+"_"+read.mappedFinal+"_"+read.startPos+"_"+read.inspected+"_"+read.prev);
			}else{
				bw.write(">"+read.runName+"_"+"___"+"_"+read.orientation+"_"+read.name);
			}
			bw.write("\n");
			for(int k=loci[i].minReadPos; k<=loci[i].maxReadPos; k++){
				bw.write(loci[i].alignment[j][k]);
			}
			bw.write("\n");
			read=read.next;
		}
		bw.flush();
		bw.close();
		System.out.println();
		loci[i].alignment=null;
		System.gc();
	}


      }catch(IOException ioe){System.out.println("<<!!ERROR main()!!>>");}
  }

  static void createAlignment(Locus locus, int maxLocusSize){

	if(locus.nReads>0){
	//first create the alignment
		locus.alignment = new char[locus.nReads*2][maxLocusSize];	//make room for unmapped mates!

		for(int x=0; x<locus.nReads*2; x++){
			for(int y=0; y<maxLocusSize; y++){
				locus.alignment[x][y]='N';		//for now fill the entire array with Ns, at end will trim the consensus sequence
			}
		}
		Read read=locus.begRead;
		int readIndex=0;
		while(read!=null){	//note that all reads get put into the alignment, even if they are not in their final location...consensus ignores the non-final reads
			locus.minReadPos=Math.min(locus.minReadPos,read.startPos);
			locus.maxReadPos=Math.max(locus.maxReadPos,read.startPos+read.seq.length());
			for(int j=0; j<read.seq.length(); j++){	//store (trimmed) sequence in alignment char by char
				locus.alignment[readIndex][j+read.startPos]=read.seq.charAt(j);
			}
			read=read.next;
			readIndex++;
		}
	}
  }

  static void getConsensus(Locus locus, int maxLocusSize){

		if(locus.nReads>0){
			locus.consensus="";
			int nA,nT,nC,nG,nN,nX;
			int totalATCG;
			int nReadsInSite;
			Read read=null;
			int j=0;
			for(int s=0; s<maxLocusSize; s++){
				if(s<locus.minReadPos || s>locus.maxReadPos){
					locus.consensus+="N";
				}else{
					nA=nT=nC=nG=nN=nX=0;
					totalATCG=0;
					nReadsInSite=0;
					read = locus.begRead;
					j=0;
					while(read!=null){
					    if(read.mappedFinal){	//skip reads that to not match well to "consensus"
						switch(locus.alignment[j][s]){
							case 'A': nA++; totalATCG++; nReadsInSite++; break;
							case 'T': nT++; totalATCG++; nReadsInSite++; break;
							case 'C': nC++; totalATCG++; nReadsInSite++; break;
							case 'G': nG++; totalATCG++; nReadsInSite++; break;
							case 'N': nN++; break;
							case '-': nX++; 	     nReadsInSite++; break;
						}
					    }
					    read=read.next;
					    j++;
					}
					//create the consensus for this base
					if(nX > (nReadsInSite*0.9) ){	//evidence that this is just a gap (deletion) relative to the reference
						locus.consensus+='-';
					}else if(totalATCG<1){	//only zero reads for this site, should be N
						locus.consensus+="N";
					}else if(nA>=totalATCG*0.9){	//>=90% agreement needed to call base
						locus.consensus+="A";
					}else if(nT>=totalATCG*0.9){	//>=90% agreement needed to call base
						locus.consensus+="T";
					}else if(nC>=totalATCG*0.9){	//>=90% agreement needed to call base
						locus.consensus+="C";
					}else if(nG>=totalATCG*0.9){	//>=90% agreement needed to call base
						locus.consensus+="G";
					}else{				//enough sequenced but to much disagreement, N
						locus.consensus+="N";
					}
				}
			}

		}
  }
  
  static void getUnmappedMates(Locus locus, int locusID, Hashtable reads, int speciesID, int maxLocusSize, String filestem, String typeName){
    try{

	if(!(new File("reads/SeparatedByLocus/"+filestem+"1.fastq")).exists()){return;}	//if the file does not exist, then no reads need to be added from this insert size
	
	BufferedReader br1 = new BufferedReader ( new InputStreamReader(new FileInputStream(   new File("reads/SeparatedByLocus/"+filestem+"1.fastq") ) ),200000*1024);	
	BufferedReader br2 = new BufferedReader ( new InputStreamReader(new FileInputStream(   new File("reads/SeparatedByLocus/"+filestem+"2.fastq") ) ),200000*1024);	

	String tempS1=br1.readLine();	//
	String tempS2=br2.readLine();	//

	Read read1;
	Read read2;

	int readIndex=0;

	//mapping missing mates
	while(tempS1!=null){
		readIndex++;		
		if(readIndex%10000==0){
			System.out.print("\r\tMerging assemblies:  "+typeName+"..."+locus.nReads+" "+readIndex+"\t");	
//			System.out.print("\rChecking read "+readIndex+" for unmapped mates...");
		}
		read1=(Read)reads.get(tempS1);
		read2=(Read)reads.get(tempS2);
		if( (read1==null && read2==null) || (read1!=null && read2!=null) || (read1!=null && read1.locus!=locusID) || (read2!=null && read2.locus!=locusID)){
			//neither was mapped or both were mapped, or neither were mapped to this locus move along...
			br1.readLine();	//sequence
			br2.readLine();	//sequence
			br1.readLine();	//+
			br2.readLine();	//+
			br1.readLine();	//quals
			br2.readLine();	//quals				
		}else if(read1!=null){
			//only read1 was mapped
			read2=new Read();
			if(read1.orientation){
				read2.seq=getRevComp(br2.readLine());	// read 1 is forward so read 2 needs to be reversed
			}else{
				read2.seq=br2.readLine();
			}
			br1.readLine();//don't need so skip
			read2.name=tempS2;
			read2.runName="Extended_"+typeName;			
			read2.orientation=!read1.orientation;
			read2.startPos=getOptimalStartPos(read2,locus,maxLocusSize);
			read2.locus=read1.locus;
			locus.minReadPos=Math.min(locus.minReadPos,read2.startPos);
			locus.maxReadPos=Math.max(locus.maxReadPos,read2.startPos+read2.seq.length());
			locus.nReads++;
			read1.mate=read2;
			read2.mate=read1;			
			locus.endRead.next=read2;
			read2.prev=locus.endRead;
			locus.endRead=read2;

			//put the sequence in its place in the alignment matrix
			for(int i=read2.startPos; i<read2.startPos+read2.seq.length(); i++){
				locus.alignment[locus.nReads-1][i]=read2.seq.charAt(i-read2.startPos);
			}
		}else if(read2!=null){
			//only read2 was mapped
			read1=new Read();
			read1.name=tempS1;
			read1.runName="Extended_"+typeName;
			if(read2.orientation){
				read1.seq=getRevComp(br1.readLine());	// read 2 is forward so read 1 does not need to be reversed
			}else{
				read1.seq=br1.readLine();
			}
			br2.readLine();	//don't need so skip
			read1.orientation=!read2.orientation;
			read1.startPos=getOptimalStartPos(read1,locus,maxLocusSize);
			locus.minReadPos=Math.min(locus.minReadPos,read1.startPos);
			locus.maxReadPos=Math.max(locus.maxReadPos,read1.startPos+read1.seq.length());
			
			read1.locus=read2.locus;
			locus.nReads++;
			read2.mate=read1;
			read1.mate=read2;

			locus.endRead.next=read1;
			read1.prev=locus.endRead;
			locus.endRead=read1;		

			//put the sequence in its place in the alignment matrix
			for(int i=read1.startPos; i<read1.startPos+read1.seq.length(); i++){
				locus.alignment[locus.nReads-1][i]=read1.seq.charAt(i-read1.startPos);
			}
		}
		
		tempS1=br1.readLine();	//+
		tempS2=br2.readLine();	//+
		tempS1=br1.readLine();	//quals
		tempS2=br2.readLine();	//quals	

		tempS1=br1.readLine();	//header
		tempS2=br2.readLine();	//header		
	}
      }catch(IOException ioe){System.out.println("<<!!ERROR main()!!>>");}	  
  }
  
  static String getRevComp(String seq){
  	String revComp="";
  	int len=seq.length();
  	for(int i=0; i<len; i++){
  		revComp+=seq.charAt(len-(i+1));
  	}
  	if(seq.length()!=revComp.length()){System.out.println("ERROR 1 IN getRevCOMP!!!");}
  
  	revComp=revComp.replace('A','Z');
  	revComp=revComp.replace('T','A');
  	revComp=revComp.replace('Z','T'); 
  	
  	revComp=revComp.replace('C','Q');
  	revComp=revComp.replace('G','C');
  	revComp=revComp.replace('Q','G');

  	//REPLACE ALL OTHER BASES WITH 'N'
  	for(int i=0; i<len; i++){
  		char c=revComp.charAt(i);
  		if(c!='-' && c!='A' && c!='T' && c!='C' && c!='G'){
  			if(len>i+1){	revComp=revComp.substring(0,i)+'N'+revComp.substring(i+1);
  			}else{		revComp=revComp.substring(0,i)+'N'; }
		}
  	}  	
  	if(revComp.length()!=len){System.out.println("ERROR 2 IN getRevCOMP!!!");}
  		
	return revComp;
  }

  static int getOptimalStartPos(Read READ, Locus locus, int maxLocusSize){
	String read=READ.seq;
	String consensus = locus.consensus;
	int nMatches=0;
	int maxMatches=0;
	int maxMatchPos=-1;
	int overlap=0;
	int maxMatchOverlap=0;
	char conChar='X';

	int gapOffset=0;

	for(int i=0; i<maxLocusSize-read.length(); i++){
		nMatches=0;
		gapOffset=0;
		overlap=0;
		for(int j=0; j<read.length(); j++){
			conChar=consensus.charAt(i+j+gapOffset);
			if(conChar=='-'){
				gapOffset++;
				j--;
			}else if(conChar=='N'){
				//do nothing
			}else if(j>=0 && read.charAt(j)==conChar){
				nMatches++;
				overlap++;
			}else{	overlap++; }	//count when mismatched as well.
						
		} 		
		if(nMatches>maxMatches){
			maxMatches=nMatches;
			maxMatchPos=i;
			maxMatchOverlap=overlap;
		}
 	}
 	
 
 	int conPos=maxMatchPos;
 	int readPos=0;
 	
 	String newRead="";
 	//NOTE THAT THE SEQUENCE ITSELF IS ADJUSTED HERE TO FIT INTO THE OPTIMAL LOCATION!!!
 	while(readPos<read.length()){
 		conChar=consensus.charAt(conPos);
 		if(conChar=='-'){
 			newRead+='-';
 		}else{
 			newRead+=read.charAt(readPos);
 			readPos++;
 		}
 		conPos++;
 	}
	READ.seq=newRead; 	
	READ.maxMatches=maxMatches;
	READ.maxMatchOverlap=maxMatchOverlap;
 	
 	return maxMatchPos;
  }

  static void refineOptimalStartLocs(Locus locus,int maxLocusSize){

  		Read read = locus.begRead;
  		while(read!=null){
  			if(!read.mappedFinal){	//previously a dud
				//remove the gaps
				String tempS="";
				for(int x=0; x<read.seq.length(); x++){
					if(read.seq.charAt(x)!='-'){
						tempS=tempS+read.seq.charAt(x);
					}
				}
				read.seq=tempS;
				//find the new optimal fit location (function puts new gaps in)
  				read.startPos=getOptimalStartPos(read,locus,maxLocusSize);
  			}
  			read=read.next;
  		}		
  }

  static void getMapProbs(Locus locus){
	int overlap;
	int nMatches;
	int gapOffset=0;
	char conChar='x';

  		Read read=locus.begRead;
  		while(read!=null){
  			nMatches=0;
  			gapOffset=0;
  			overlap=0;
			for(int j=0; j<read.seq.length(); j++){
				conChar=locus.consensus.charAt(read.startPos+j+gapOffset);

				if(conChar=='-'){
					//don't count this base...
				}else if(j>=0 && read.seq.charAt(j)==conChar){
					nMatches++;
					overlap++;
				}else{
					if(conChar!='N'){
						overlap++;
					}
				}
			}
			read.maxMatches=nMatches;
			read.maxMatchOverlap=overlap;
			read.mapProb=nMatches/(double)overlap;
  			read=read.next;
  		}
  }

  static int updateMappedFinal(Locus locus){
	int improvement=0;

  		Read read = locus.begRead;
  		while(read!=null){
  			if(!read.mappedFinal && read.maxMatchOverlap>=25 && read.maxMatches>=read.maxMatchOverlap*0.9){
  				improvement++;
  				read.mappedFinal=true;
  				if(read.startPos<locus.minReadPos){
  					locus.minReadPos=read.startPos;
  				}
  				
  				if(read.startPos+read.seq.length()>locus.maxReadPos){
  					locus.maxReadPos=read.startPos+read.seq.length();	//note that this does not account for gaps...
  				}
  			}
  			read=read.next;
  		}
 	return improvement;
  }

  static boolean checkIntegrity(Locus locus){
   	Read read = locus.begRead;
	int count=0;
  	while(read!=null){
  		count++;
  		if(read.next!=null && read.next.prev!=read){
  			System.out.println("Locus failed the integrity check1 (read.next.prev!=read)");
			displayReadLinks(locus);
			return false;
  		}
  		if(read.prev!=null && read.prev.next!=read){
  			System.out.println("Locus failed the integrity check2 (read.prev.next!=read)");  					
			displayReadLinks(locus);
			return false;
  		}
  		
  		read=read.next;
  	}
  		
  	if(count!=locus.nReads){
  		System.out.println("Locus failed the integrity check3 (count!=locus)"); 
		displayReadLinks(locus);
  		return false;
  	}

	read = locus.endRead;
	count=0;
  	while(read!=null){
  		count++;
  		if(read.next!=null && read.next.prev!=read){
  			System.out.println("Locus failed the integrity check4 (read.next.prev!=read)");
			displayReadLinks(locus);
			return false;
  		}
  		if(read.prev!=null && read.prev.next!=read){
  			System.out.println("Locus failed the integrity check5 (read.prev.next!=read)");
			displayReadLinks(locus);
			return false;
  		}
  		read=read.prev;
  	}	

  	if(count!=locus.nReads){
  		System.out.println("Locus failed the integrity check5 (count!=locus)"); 
		 displayReadLinks(locus);
  		return false;
  	} 
  	return true;	
  }

  static void displayReadLinks(Locus locus){
  	Read read=locus.begRead;
  	while(read!=null){
  		System.out.println(read.prev+"<--"+read+"-->"+read.next);
		read=read.next;
  	}
  }

  static void removeDuds(Locus locus){
	int readCount=0;
  	int dudCount=0;

	  	Read read=locus.endRead;
	  	readCount+=locus.nReads;
	  	while(read!=null && !read.mappedFinal){
  			read=read.prev;
  			locus.nReads--;
			dudCount++;
	  	}
	  	if(read!=null && read.next!=null){
	  		read.next.prev=null;
	  		read.next=null;
	  		locus.endRead=read;
	  	}
  }

/*
  static void cleanUpAlignment(Locus locus, int maxErrorsAllowed[], int maxLocusSize){
	System.out.println("Identifying site occupancy...");
	getOccupiedArray(locus, maxLocusSize);
	System.out.println("Cleaning up sequencing errors...");
	char c='x';

	  	//**Note that this method assumes that an alignment was just created (works off alignment not reads)...
		int nReads=0;
		int nA=0;
		int nT=0;
		int nC=0;
		int nG=0;
		int maxCount=0;
		char maxBase = '~';
		int errors=0;
	  	for(int j=locus.minReadPos; j<=locus.maxReadPos; j++){
			nReads=0;
			nA=0;
			nT=0;
			nC=0;
			nG=0;
			maxCount=0;
			maxBase = '~';
			errors=0;			
			for(int r=0; r<locus.nReads; r++){
				switch(locus.alignment[r][j]){
					case 'A': nA++; nReads++; break;
					case 'T': nT++; nReads++; break;
					case 'C': nC++; nReads++; break;
					case 'G': nG++; nReads++; break;
					case 'N': if(locus.occupied[r][j]){ nReads++; }
				}
			}
			maxCount=0;
			if(nA>maxCount){ maxCount=nA; maxBase='A'; }
			if(nT>maxCount){ maxCount=nT; maxBase='T'; }
			if(nC>maxCount){ maxCount=nC; maxBase='C'; }
			if(nG>maxCount){ maxCount=nG; maxBase='G'; }		
			errors=nReads-maxCount;
			if(errors <= maxErrorsAllowed[nReads]){	//these thresholds are precomputed in the main function, if below min reads needed and should ignore site, -1 exists in maxErrorsAllowed
				//need to clean up this site!
				for(int r=0; r<locus.nReads; r++){
					c=locus.alignment[r][j];
					if( (c=='A' || c=='T' || c=='C' || c=='G' || (c=='N' && locus.occupied[r][j])) && c!=maxBase){
						locus.alignment[r][j]=maxBase;
					}
				}
			}	
			
	  	}
	//System.out.println("\n\t"+totalCleanedUp+" sequencing errors repaired of "+totalBases+" total bases mapped (proportion cleaned = "+(totalCleanedUp/(double)totalBases)+")");
  }
*/

  static void sortReadsInAssembly(Locus locus){

  		Read read = locus.begRead;

		while(read!=null){
			read.inspected=false;
			read=read.next;
		}
		read = locus.begRead;

  		while(read!=null && read.next!=null && !read.next.inspected){		//once an inspected read.next is reached, should be in the dud section at the end of the read list

			read.next.inspected=true;
  			if(!read.next.mappedFinal){		//need to move to the end if not already there
				if(read==locus.begRead){	//currently at the beginning...
					locus.begRead=read.next;
					read.next.prev=null;
					read.next=null;
					read.prev=locus.endRead;
					locus.endRead.next=read;
					locus.endRead=read;
					read=locus.begRead;		//reset read so that it is at the new start of the list
				}else if(read.next!=locus.endRead){	//somewhere in the middle	

					Read tempRead=read.next;
					tempRead.next.prev=read;
					read.next=tempRead.next;
					tempRead.prev=locus.endRead;
					locus.endRead.next=tempRead;
					tempRead.next=null;
					locus.endRead=tempRead;
				}
  			}else if(read.startPos>read.next.startPos){

  				//need to move the read.next up in the list
		  		Read tempRead=read;		//this pointer will move until the correct location is found

				Read tempRead2=read.next;	//this pointer will remain pointing to the read that will be moved
		  		
				while(tempRead!=null && tempRead.startPos > tempRead2.startPos){	//move up until the correct location is found
					tempRead=tempRead.prev;					
				}

				read.next=read.next.next;				//connect around read that is being moved

				if(locus.begRead==read){				//read to be moved will be moved from second to first
					tempRead2.next=read;
					tempRead2.prev=null;
					read.prev=tempRead2;
					locus.begRead=tempRead2;
				}else if(tempRead==null){				//needs to be moved to the first in the list
					tempRead2.next=locus.begRead;			//connect the moved node to the previous firstNode in the list
					tempRead2.next.prev=tempRead2;			//connect the other direction
					tempRead2.prev=null;
					locus.begRead=tempRead2;			//connect the begRead pointer to this moved read
				}else{							//if not move to the beginning, insert it in the new location
					tempRead2.next=tempRead.next;			
					tempRead2.prev=tempRead;
					tempRead.next.prev=tempRead2;
					tempRead.next=tempRead2;
				}
				
				if(read.next==null){					//the moved node was previously at the end of the list
					locus.endRead=read;				//point the endRead pointer to the new end
				}else{
					read.next.prev=read;				//create the reverse connection if appropriate
				}

  			}else{
  				//only advance if the next one was just inspected and found ok
  				read=read.next;
			}			
			
			if(!checkIntegrity(locus)){System.out.println("BAD INTEGRITY"); return;}

  		}
  }
  
  static void writeNReadsFixed(Locus locus, int speciesID, boolean append){
  	try{
  		BufferedWriter bw = new BufferedWriter ( new OutputStreamWriter(new FileOutputStream(   new File("Assemblies/"+speciesID+"/nReadsFixed.txt"),append ) ));	//to store summary of nReadsMapped...
		bw.write(locus.nReads+"\n");
		bw.flush();
		bw.close();
  	}catch(IOException ioe){System.out.println("IOE in writeFinalReadsMapped!");}
  }

  static void getOccupiedArray(Locus locus, int maxLocusSize){	//assumes that the alignment is ready to go
  	locus.occupied=new boolean[locus.nReads][maxLocusSize];
	for(int i=0; i<locus.nReads; i++){
		for(int s=0; s<maxLocusSize; s++){
			locus.occupied[i][s]=true;
		}
	}
	//switch the NS before the first nonN
  	for(int i=0; i<locus.nReads; i++){
		for(int s=0; s<maxLocusSize; s++){
  			if(locus.alignment[i][s]=='N'){
  				locus.occupied[i][s]=false;
  			}else{
  				break;
  			}
  		}
	}
	//switch the ns after the last nonN
	for(int i=0; i<locus.nReads; i++){
		for(int s=maxLocusSize-1; s>=0; s--){
  			if(locus.alignment[i][s]=='N'){
  				locus.occupied[i][s]=false;
  			}else{
  				break;
  			}
  		}
	}
	
  }


////////////////////////////////////////////////////
////////////////////////////////////////////////////

/*	python version
def pbinomial(q, n, p):
    result=0
    for x in range(0,q+1):
        result+=dbinomial(x,n,p,False)
    return(result)
*/

//	java version
  static double pbinomial(double q, int n, double p){
  	double result=0;
  	for(int x=0; x<q+1; x++){
  		result+=dbinomial(x,n,p,false);
  	}
  	return result;
  }

/*	python version
def dbinomial(x, n, p, returnLog):
    result = log_factorial(n) - (log_factorial(x)+log_factorial(n-x)) + (x * log(p)) + ((n-x) * log(1-p))
    if returnLog:
        return result
    else:
        return exp(result)
*/

//	java version
  static double dbinomial(int x, int n, double p, boolean returnLog){
  	double result = log_factorial(n) - (log_factorial(x)+log_factorial(n-x)) + (x * Math.log(p)) + ((n-x) * Math.log(1-p));
	if(returnLog){
		return result;
	}else{
		return Math.exp(result);
	}
  }

/*
#this function only applies to the case of four categories
def dmultinomial(n, xs, ps, returnLog):
    result = log_factorial(n) - (log_factorial(xs[0])+log_factorial(xs[1])+log_factorial(xs[2])+log_factorial(xs[3])) + (xs[0] * log(ps[0])) + (xs[1] * log(ps[1])) + (xs[2] * log(ps[2])) + (xs[3] * log(ps[3]))

    if returnLog:
        return result
    else:
        return exp(result)
*/

/*	python version
def log_factorial(x):
    #"""Returns the logarithm of x!
    #Also accepts lists and NumPy arrays in place of x."""
    return gammaln(x+1)    
*/

//	java version
  static double log_factorial(int x){
  	//"Returns the logarithm of x!
  	return gammaln(x+1);
  }

/*	python version
def gammaln(n):
    #"""Logarithm of Euler's gamma function for discrete values."""
    if n < 1:
        return float('inf')
    if n < 3:
        return 0.0
    c = [76.18009172947146, -86.50532032941677, \
         24.01409824083091, -1.231739572450155, \
         0.001208650973866179, -0.5395239384953 * 0.00001]
    x, y = float(n), float(n)
    tm = x + 5.5
    tm -= (x + 0.5) * log(tm)
    se = 1.0000000000000190015
    for j in range(6):
        y += 1.0
        se += c[j] / y
    return -tm + log(2.5066282746310005 * se / x)
*/

//	java version
  static double gammaln(int n){
    	//Logarithm of Euler's gamma function for discrete values. 	
  	if(n<1){
  		return Double.POSITIVE_INFINITY;
  	}else if(n<3){
  		return 0;
  	}else{
  		double c[]={76.18009172947146, -86.50532032941677, 24.01409824083091, -1.231739572450155, 0.001208650973866179, -0.5395239384953 * 0.00001};
  		double x = n;
  		double y = n;
  		double tm = x + 5.5;
  		tm -= (x + 0.5) * Math.log(tm);
  		double se =  1.0000000000000190015;
  		for(int j=0; j<6; j++){
  			y += 1.0;
  			se += c[j] / y;
  		}
  		return -tm + Math.log(2.5066282746310005 * se / x);
  	}
  
  
  }


///////////////////////////////////////////////////////
///////////////////////////////////////////////////////
}

class Locus{
	int nReads=0;
	Read begRead=null;
	Read endRead=null;
	char alignment[][];
	boolean occupied[][];
	String consensus="";
	int minReadPos=6000;
	int maxReadPos=0;
}

class Read{
	int locus=-1;
	String runName="";
	String name;
	boolean orientation;
	Read next=null;
	Read prev=null;
	String seq="";			//this is the read sequence, oriented
	int startPos=-1;
	Read mate=null;
	int maxMatches=-1;
	int maxMatchOverlap=-1;
	boolean mappedFinal=false;
	boolean inspected=false;
	double mapProb=0;
}
