import java.io.*;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.*;

////////////////////////////////////
//THIS PROGRAM TAKES A CONSENSUS SEQUENCE FROM AN ALIGNMENT IN FASTA FORMAT, DEGREE OF CONSENSUS (E.G. STRICT OR MAJORITY RULE) IS DEFINED BY THE USER
//NOTE THAT POLYMORPHISMS ARE TREATED USING AMBIGUITY CODES BUT NOT PHASED OUT

public class GetAlignmentConsensus{
  public static void main(String[] args){
      try{

	String specAbbrev[] = {"A","H","G","X","Z","M","K","L","P","S"};


	for(int speciesIndex=1; speciesIndex<11; speciesIndex++){
		BufferedWriter bw = new BufferedWriter ( new OutputStreamWriter(new FileOutputStream(   new File("Assemblies/"+speciesIndex+"/ReducedReadAlignments/ConsensusSequences_"+specAbbrev[speciesIndex-1]+".fasta") ) ));

		for(int locusIndex=1; locusIndex<513; locusIndex++){
			System.out.print("\r"+specAbbrev[speciesIndex-1]+"\t"+locusIndex);
			BufferedReader br = new BufferedReader ( new InputStreamReader(new FileInputStream(   new File("Assemblies/"+speciesIndex+"/ReducedReadAlignments/L"+locusIndex+specAbbrev[speciesIndex-1]+".fasta") ) ));

			int nSeqs=0;
			int nSites=0;
		
			//first identify the size of the matrix
			String header=br.readLine();
			String sequence;
			while(header!=null){
				sequence=br.readLine();
				nSeqs++;
				if(nSeqs==1){ nSites=sequence.length(); }
				header=br.readLine();
			}	
			br.close();
		
			//now reopen the file and load the matrix
			br = new BufferedReader ( new InputStreamReader(new FileInputStream(   new File("Assemblies/"+speciesIndex+"/ReducedReadAlignments/L"+locusIndex+specAbbrev[speciesIndex-1]+".fasta") ) ));
			char alignment[][] = new char[nSeqs][nSites];
			int begSeq[] = new int[nSeqs];
			int endSeq[] = new int[nSeqs];
			Arrays.fill(begSeq,-1);
			Arrays.fill(endSeq,-1);
			
			for(int i=0; i<nSeqs; i++){
				header=br.readLine();
				sequence=br.readLine();
				for(int j=0; j<nSites; j++){
					alignment[i][j]=sequence.charAt(j);
					if(alignment[i][j]!='-'){ begSeq[i]=j; break;}
				}
			}
			for(int i=0; i<nSeqs; i++){
				for(int j=nSites-1; j>=0; j--){
					if(alignment[i][j]!='-'){ endSeq[i]=j; break;}
				}
			}


		
			//now write the consensus sequence to a file
			//String consensus="";
			bw.write(">L"+locusIndex+specAbbrev[speciesIndex-1]+"\n");
			int nA,nT,nC,nG,nN,nATCG;
			int nGap=0;		
			char maxBase;
			int maxCount;
			for(int j=0; j<nSites; j++){
				nA=nT=nC=nG=nN=0;	
				for(int i=0; i<nSeqs; i++){
					if(j>=begSeq[i] && j<=endSeq[i]){
						switch(alignment[i][j]){
							case 'A': nA++; break;
							case 'T': nT++; break;
							case 'C': nC++; break;
							case 'G': nG++; break;
							case 'N': nN++; break;
							case '-': nGap++; break;
						}
					}
				}
				
				//identify the maximum
				maxBase='A';
				maxCount=nA;
				nATCG=nA+nT+nC+nG;
				if(nT>maxCount){	maxBase='T'; maxCount=nT;	}
				if(nC>maxCount){	maxBase='C'; maxCount=nC;	}
				if(nG>maxCount){	maxBase='G'; maxCount=nG;	}
				if(nGap>maxCount){	maxBase='-'; maxCount=nGap;	}
				
				if(maxBase!='-' && maxCount>1){
					if(maxBase=='-'){
						//do nothing since this is a gap
					}else if(nATCG<=1 || maxCount/(double)nATCG<=0.5){
						bw.write('N');
					}else{
						bw.write(maxBase);
					}
				}
			}	
			bw.write("\n");
			bw.flush();
		}
	}
	System.out.println();
	
      }catch(IOException ioe){System.out.println("<<!!ERROR main()!!>>");}
  }


}
