import java.io.*;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.*;

////////////////////////////////////
//THIS PROGRAM TAKES A CONSENSUS SEQUENCE FROM AN ASSEMBLY IN FASTA FORMAT, DEGREE OF CONSENSUS (E.G. STRICT OR MAJORITY RULE) IS DEFINED BY THE USER
// if nonNs < minCoverage at site, N
// else if one allele with sequencing error, maxBase
// else if two alleles with sequencing error, ambiguity code
// else N
//NOTE THAT POLYMORPHISMS ARE TREATED USING AMBIGUITY CODES BUT NOT PHASED OUT

public class GetAssemblyConsensus{
  public static void main(String[] args){
      try{

	String specAbbrev[] = {"A","H","G","X","Z","M","K","L","P","S"};
      
      	String folder = args[0];

	int minCoverage = 5;
	boolean doLocus[] = new boolean[512];	
	if(args.length>1){
		doLocus[(Integer.parseInt(args[1])-1)]=true;	//if user specifies a single locus, just do that locus
	}else{							//if user does not specify, then to all loci that have at least one species representing
		for(int i=0; i<512; i++){
			for(int j=0; j<10; j++){
				if(new File(folder+"/L"+(i+1)+specAbbrev[j]+".fasta").exists()){	//if user does not specify, then to all loci that have at least one species representing
					doLocus[i]=true;
					break;
				}
			}
		} 
	}				
      
      	BufferedWriter bw;
      	for(int locus=1; locus<513; locus++){
	      	if(doLocus[locus-1]){
		      	bw = new BufferedWriter ( new OutputStreamWriter(new FileOutputStream(   new File(folder+"/L"+locus+".fasta") ) ));	//this file will contain reference sequences plus consensus from all species for which assemblies exist in the folder specified
		
			//first gather the reference sequences for this locus
			for(int s=0; s<4; s++){
				BufferedReader brP = new BufferedReader ( new InputStreamReader(new FileInputStream(   new File(folder+"/ProbeRegionSeqs"+specAbbrev[s]+".fasta") ) ));	
				String tempS=brP.readLine();	//e.g. >L100A
				while(tempS!=null){
					if(tempS.startsWith(">L"+locus+specAbbrev[s])){
						bw.write(">ProbeRegion"+locus+specAbbrev[s]+"\n");
						bw.write(brP.readLine()+"\n");
					}
					tempS=brP.readLine();
				}			
				brP.close();
			}

		      	//loop through all potential species, gather all species for which there is a file in the given folder
			System.out.print(locus+"\t");
		      	for(int s=0; s<10; s++){
		      		String assemblyFile = folder+"/L"+locus+specAbbrev[s]+".fasta";
				if(new File(assemblyFile).exists()){
					System.out.print(specAbbrev[s]);
					BufferedReader br = new BufferedReader ( new InputStreamReader(new FileInputStream(   new File(assemblyFile) ) ));
		
		
					int nSeqs=0;
					int nSites=0;
				
					//first identify the size of the matrix
					String header=br.readLine();
					String sequence;
					while(header!=null){
						sequence=br.readLine();
						nSeqs++;
						if(nSeqs==1){ nSites=sequence.length(); }
						header=br.readLine();
					}	
					br.close();
						
					//now reopen the file and load the matrix
					br = new BufferedReader ( new InputStreamReader(new FileInputStream(   new File(assemblyFile) ) ));
					char alignment[][] = new char[nSeqs][nSites];
					int begSeq[] = new int[nSeqs];
					int endSeq[] = new int[nSeqs];
					Arrays.fill(begSeq,-1);
					Arrays.fill(endSeq,-1);
							
					//first store the sequences in the alignment
					for(int i=0; i<nSeqs; i++){
						header=br.readLine();
						sequence=br.readLine();
						for(int j=0; j<nSites; j++){
							alignment[i][j]=sequence.charAt(j);
						}
					}
				
					//now find the first base for each sequence
					for(int i=0; i<nSeqs; i++){
						for(int j=0; j<nSites; j++){
							if(alignment[i][j]!='-'){begSeq[i]=j; break;}
						}
					}
					
					for(int i=0; i<nSeqs; i++){
						for(int j=nSites-1; j>=0; j--){
							if(alignment[i][j]!='-'){ endSeq[i]=j; break;}
						}
					}
						
					//now write the consensus sequence to a file
					bw.write(">L"+locus+specAbbrev[s]+"\n");
					int nA,nT,nC,nG,nN,nATCG,nATCGGap;
					int nGap=0;		
					char maxBase,maxBase2;
					int maxCount,maxCount2;
					
					String cs="";	//consensus sequence
					
					for(int j=0; j<nSites; j++){
						nA=nT=nC=nG=nN=nGap=0;	
						for(int i=0; i<nSeqs; i++){
							if(j>=begSeq[i] && j<=endSeq[i]){
								switch(alignment[i][j]){
									case 'A': nA++; break;
									case 'T': nT++; break;
									case 'C': nC++; break;
									case 'G': nG++; break;
									case 'N': nN++; break;
									case '-': nGap++; break;
								}
							}
						}
								
						//identify the maximum
						maxBase='X';
						maxCount=0;
						maxBase2='X';
						maxCount2=0;
						nATCG=nA+nT+nC+nG;	
						nATCGGap=nA+nT+nC+nG+nGap;
				
						//find most common base
						if(nA>=maxCount){	maxBase='A'; maxCount=nA;	}	
						if(nT>=maxCount){	maxBase='T'; maxCount=nT;	}
						if(nC>=maxCount){	maxBase='C'; maxCount=nC;	}
						if(nG>=maxCount){	maxBase='G'; maxCount=nG;	}
						if(nGap>=maxCount){	maxBase='-'; maxCount=nGap;	}
				
						//find second most common base		
						if(maxBase!='A' && nA>maxCount2){	maxBase2='A'; maxCount2=nA;}	
						if(maxBase!='T' && nT>maxCount2){	maxBase2='T'; maxCount2=nT;}	
						if(maxBase!='C' && nC>maxCount2){	maxBase2='C'; maxCount2=nC;}	
						if(maxBase!='G' && nG>maxCount2){	maxBase2='G'; maxCount2=nG;}	
						if(maxBase!='-' && nGap>maxCount2){	maxBase2='-'; maxCount2=nGap;}
				
				
				
						if(nATCG<minCoverage){ 	if(maxBase!='-'){						cs=cs+"N"; }		//not enough coverage to call the base...write nothing if gap is most common
						}else if(maxCount2==0){	if(maxBase!='-'){						cs=cs+maxBase; }	//enough coverage and all called bases are identical...note that nothing is written if all gaps!
						}else if(isTwoAlleles(maxCount,maxCount2,nATCGGap,j)){								
							if(	 (maxBase=='A' && maxBase2=='T') || (maxBase=='T' && maxBase2=='A')){	cs=cs+'W';
							}else if((maxBase=='A' && maxBase2=='C') || (maxBase=='C' && maxBase2=='A')){	cs=cs+'M';
							}else if((maxBase=='A' && maxBase2=='G') || (maxBase=='G' && maxBase2=='A')){	cs=cs+'R';
							}else if((maxBase=='A' && maxBase2=='-') || (maxBase=='-' && maxBase2=='A')){	cs=cs+'a';	//lower case denotes indel...to be used for polymorphism calculator later
							}else if((maxBase=='T' && maxBase2=='C') || (maxBase=='C' && maxBase2=='T')){	cs=cs+'Y';
							}else if((maxBase=='T' && maxBase2=='G') || (maxBase=='G' && maxBase2=='T')){	cs=cs+'K';
							}else if((maxBase=='T' && maxBase2=='-') || (maxBase=='-' && maxBase2=='T')){	cs=cs+'t';
							}else if((maxBase=='C' && maxBase2=='G') || (maxBase=='G' && maxBase2=='C')){	cs=cs+'S';
							}else if((maxBase=='C' && maxBase2=='-') || (maxBase=='-' && maxBase2=='C')){	cs=cs+'c';
							}else if((maxBase=='G' && maxBase2=='-') || (maxBase=='-' && maxBase2=='G')){	cs=cs+'g';
							}else{ 										cs=cs+'N';	System.out.println("not sure what happend here!"); }		
						}else if(maxCount>maxCount2){								cs=cs+maxBase;			//majority rule base call if no tie
				//		}else if(isOneAllele(maxCount,nATCGGap,j)){	if(maxBase!='-'){ 			cs=cs+maxBase; }		//evidence for one allele...write nothing if gap is most common
						}else{											cs=cs+'N';		}	
					}
					//NOW TRIM OFF THE N'S ON THE ENDS...
					for(int i=0; i<cs.length(); i++){
						if(cs.charAt(i)!='N'){
							for(int j=cs.length()-1; j>=0; j--){
								if(cs.charAt(j)!='N'){
									bw.write(cs.substring(i,j+1)+"\n");
									break;
								}
							}
							break;
						}
						if(i==cs.length()-1){bw.write(cs+"\n");}
					}
				}
			}
			System.out.println();
			bw.flush();
			bw.close();
		}
	}
      }catch(IOException ioe){System.out.println("<<!!ERROR main()!!>>");}
  }

  static boolean isOneAllele(int nX, int total, int j){
  	return (1-pbinomial(total-nX,total,0.05))>=0.01;						//allow for 5% sequencing error
  }
  
  static boolean isTwoAlleles(int nX, int nY, int total, int j){
  	return (1-pbinomial(nX,nX+nY,0.5))>=0.01 && (1-pbinomial(total-(nX+nY),total,0.05))>=0.01;	//acceptable ratio for two alleles and acceptable amount of sequencing error
  }

  static double pbinomial(double q, int n, double p){
  	double result=0;
  	for(int x=0; x<q+1; x++){
  		result+=dbinomial(x,n,p,false);
  	}
	if(result>1){
		return 1;
	}else{
  		return result;
 	}
  }
  
  static double dbinomial(int x, int n, double p, boolean returnLog){
  	double result = log_factorial(n) - (log_factorial(x)+log_factorial(n-x)) + (x * Math.log(p)) + ((n-x) * Math.log(1-p));
	if(returnLog){
		return result;
	}else{
		return Math.exp(result);
	}
  }
  
  static double log_factorial(int x){
  	//"Returns the logarithm of x!
  	return gammaln(x+1);
  }

  static double gammaln(int n){
    	//Logarithm of Euler's gamma function for discrete values. 	
  	if(n<1){
  		return Double.POSITIVE_INFINITY;
  	}else if(n<3){
  		return 0;
  	}else{
  		double c[]={76.18009172947146, -86.50532032941677, 24.01409824083091, -1.231739572450155, 0.001208650973866179, -0.5395239384953 * 0.00001};
  		double x = n;
  		double y = n;
  		double tm = x + 5.5;
  		tm -= (x + 0.5) * Math.log(tm);
  		double se =  1.0000000000000190015;
  		for(int j=0; j<6; j++){
  			y += 1.0;
  			se += c[j] / y;
  		}
  		return -tm + Math.log(2.5066282746310005 * se / x);
  	}
  
  
  }


}
