import java.io.*;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.*;

////////////////////////////////////
//THIS PROGRAM TAKES AN ALIGNMENT AND REDUCES IT TO JUST THE NON-REDUNDANT SEQUENCES (TWO IDENTICAL SEQUENCES BECOME ONE IDENTICAL SEQUENCE)
//ALSO TAKES INTO ACCOUNT MAPPED/EXTENDED AND SEQUENCE TYPE HISEQ500, ETC...
public class ReduceAlignment3{
  public static void main(String[] args){
      try{

	/////////////PRECOMPUTE MAXIMUM ERRORS ALLOWED USING PBINOM////////
	System.out.println("Precalculating binomial probabilities...");
	int maxErrorsAllowed[] = new int[30000];
	maxErrorsAllowed[0]=-1;
	int tempMax=1;
	for(int i=1; i<30000; i++){			//maximum of 30,000 coverage per site allowed
		while(1-pbinomial(tempMax,i,0.01)>0.01){	tempMax++;	}	//error probability of 1% assumed, use alpha equals 0.01...
		tempMax--;	//since the last one failed, subtract it off
		if(tempMax<1){	maxErrorsAllowed[i]=-1;		//trigger to skip site since not enough reads were sequenced to tell anything
		}else{		maxErrorsAllowed[i]=tempMax;	}

		//if(i<150){System.out.println(i+"\t"+maxErrorsAllowed[i]);}
	}
	///////////////////////////////////////////////////////////////////

	int speciesIndex=Integer.parseInt(args[0]);
	char speciesID=args[1].charAt(0);

	BufferedWriter bw1 = new BufferedWriter ( new OutputStreamWriter(new FileOutputStream(   new File("Assemblies/"+speciesIndex+"/nReadsBeforeReduction_"+speciesIndex+".txt") ) ));
	BufferedWriter bw2 = new BufferedWriter ( new OutputStreamWriter(new FileOutputStream(   new File("Assemblies/"+speciesIndex+"/nReadsAfterReduction_"+speciesIndex+".txt") ) ));
	BufferedWriter bwL = new BufferedWriter ( new OutputStreamWriter(new FileOutputStream(   new File("Assemblies/"+speciesIndex+"/lengthsAfterReduction_"+speciesIndex+".txt") ) ));


	for(int locusIndex=1; locusIndex<513; locusIndex++){
		System.out.print("\rRefining assembly of locus "+locusIndex);
	
		BufferedReader br = new BufferedReader ( new InputStreamReader(new FileInputStream(   new File("Assemblies/"+speciesIndex+"/ReadAlignments/L"+locusIndex+speciesID+".fasta") ) ));
		BufferedWriter bw = new BufferedWriter ( new OutputStreamWriter(new FileOutputStream(   new File("Assemblies/"+speciesIndex+"/ReducedReadAlignments/L"+locusIndex+speciesID+".fasta") ) ));

		int nSites=-1;

		//////
		//STORE ALL OF THE PAIR INFORMATION
		Hashtable pairs = new Hashtable(2000000);	//HASH TABLE USED TO STORE PAIR INFORMATION
		String tempS=br.readLine();	//get the header	e.g.: >Extended_Hi900_894_true_@HWI-ST1035:54:D03ERACXX:4:1105:21361:20274/1
		String word[];
		String name="";
		Pair pair;
		char side;
		String MorE="";

		while(tempS!=null){
			word=tempS.split("@");
			name="@"+word[1];
			word=word[0].split("_");
			MorE=word[0].substring(1);
			if(name.charAt(name.length()-2)=='/'){
				side=name.charAt(name.length()-1);		// e.g. '1' or '2'
				name=name.substring(0,name.length()-2);
			}else{
				side='M';
			}
			pair=(Pair)pairs.get(name);

			if(pair==null || (pair!=null && side!='M')){	// pair exists and thought to be merged, but here is another read so must be an issue to resolve
									//hack to keep unpaired reads from contributing if merged version is already included...
				if(pair==null){	//have not observed a read from this pair before...
					pair=new Pair();
					pair.name=name;
					pair.read1=br.readLine();
					pair.library=word[1];
					pair.side1=side;
					setBegPos(pair);
					if(side=='M'){
						pair.side2='M';
						setEndPos(pair);
					}
					pair.MorE1=MorE;
					pairs.put(name,pair);
					pair.timesObserved=1;
				}else{		//have already observed a read from this pair...
					pair.read2=br.readLine();
					pair.side2=side;
					setEndPos(pair);		//also computes the insert size...
					pair.MorE2=MorE;
					pair.timesObserved++;
				}
			
				if(pair.endPos==-1){setEndPos(pair);}
			}else{
				br.readLine();
			}
			
			nSites=pair.read1.length();
			tempS=br.readLine();	//get next header
		}
		br.close();

		////
		//DECIDE ON WHICH PAIR WILL BE STORED	
		Hashtable dups = new Hashtable(2000000);	//HASH TABLE USED TO STORE PCR DUPLICATES

	
		br = new BufferedReader ( new InputStreamReader(new FileInputStream(   new File("Assemblies/"+speciesIndex+"/ReadAlignments/L"+locusIndex+speciesID+".fasta") ) ));
		name=br.readLine();	//get the header e.g.: >Extended_Hi900_894_true_@HWI-ST1035:54:D03ERACXX:4:1105:21361:20274/1
		int readIndex=0;
		Dup dup;
		String key;
				
		int safeEnd=-1;

		//COLLAPSE ALL READS TO A UNIQUE SET (KEEP TRACK OF THE NUMBER OF TIMES EACH ONE WAS OBSERVED...
		while(name!=null){
			readIndex++;		//this will be used to store total number of read originally mapped
			word=name.split("@");
			name="@"+word[1];
			if(name.charAt(name.length()-2)=='/'){
				name=name.substring(0,name.length()-2);
			}
			pair=(Pair)pairs.get(name);	//note that this should certainly have been seen before...
			safeEnd=Math.min(pair.read1.length(),pair.begPos+100);
			if(pair.read2==null){
				key=pair.begPos+"_"+pair.endPos+"_"+pair.library.substring(2)+"_"+pair.read1.substring(pair.begPos,safeEnd)+"_"+"?";	//reverse read unknown
			}else{
				key=pair.begPos+"_"+pair.endPos+"_"+pair.library.substring(2)+"_"+pair.read1.substring(pair.begPos,safeEnd)+"_"+pair.read2.substring(pair.endPos-99,pair.endPos+1);			
			}
			//note library Hi and Mi are trimmed off so Hi500==Mi500 since they were the same library  OUTER 100 BP ARE TAKEN FROM INSERT
			//TRIMMING SHOULD HAVE BEEN DONE CONSISTANTLY, SO THAT WILL NOT PRESENT A PROBLEM (EXCEPT HI SEQ VS MI SEQ IF HISEQ IS TRIMMED)
			dup=(Dup)dups.get(key);	//may not have been put into the dups list
			if(dup==null){
				dup = new Dup();
				pair.dup=dup;
				dup.bestPair=pair;
				pair.chosen=true;
				dup.count=1;
				dups.put(key,dup);
				br.readLine();	//ignore read line
			}else{
				dup.count++;
				if(getReadLength(br.readLine())>Math.max(getReadLength(dup.bestPair.read1),getReadLength(dup.bestPair.read2))){	//make sure that the least trimmed read gets used (e.g. MiSeq over HiSeq)
					dup.bestPair.dup=null;
					dup.bestPair.chosen=false;	//make the old best not the chosen one
					dup.bestPair=pair;
					dup.bestPair.dup=dup;
					dup.bestPair.chosen=true;		//make the new best the chosen one
				}
			}
			pair.dup=dup;
			name=br.readLine();
		}
		br.close();

		//count up the number of sequences that will appear in the condensed alignment...
		int nSeqs=0;
		for (Enumeration e = dups.elements() ; e.hasMoreElements() ;) {
			dup=(Dup)e.nextElement();
			if(dup.bestPair.read2==null){ nSeqs+=1; }else{ nSeqs+=2; }			

			if(dup.bestPair.name.equals("@HWI-ST1035:54:D03ERACXX:4:1306:7211:35647") || dup.bestPair.name.equals("@HWI-ST1035:54:D03ERACXX:4:2305:21029:59332")){

			}
     		}

		//open once more and build the character matrix
		br = new BufferedReader ( new InputStreamReader(new FileInputStream(   new File("Assemblies/"+speciesIndex+"/ReadAlignments/L"+locusIndex+speciesID+".fasta") ) ));
		name=br.readLine();	//get the header e.g.: >Extended_Hi900_894_true_@HWI-ST1035:54:D03ERACXX:4:1105:21361:20274/1
		String sequence="";
		int totalReads=0;
		int counter=0;
		String seqHeaders[] = new String[nSeqs];
		char alignment[][] = new char[nSeqs][nSites];
		int begPos[]=new int[nSeqs];
		int endPos[]=new int[nSeqs];
		ReadIndex begReadIndex=null;
		ReadIndex currReadIndex=null;

		while(name!=null){
			sequence=br.readLine();
			totalReads++;
			word=name.split("@");
			name="@"+word[1];
			side='M';
			if(name.charAt(name.length()-2)=='/'){		//not overlapped...
				side=name.charAt(name.length()-1);
				name=name.substring(0,name.length()-2);
			}
			pair=(Pair)pairs.get(name);	//note that this should certainly have been seen before...
			if(pair.chosen && (side==pair.side1 || side==pair.side2)){	//this read is from a pair that was chosen to be written to the reduced alignment
				seqHeaders[counter]=">seq"+counter+"_"+pair.dup.count+"_"+pair.library+"_"+pair.insertSize+"_"+name+"/"+side;				//e.g. >seq94_E_8_HiSeq500_375_@HWI-ST..../1

				for(int j=0; j<nSites; j++){
					alignment[counter][j]=sequence.charAt(j);
				}

				////////////////////////////////////
				//this section obtains information needed to adjust the order of the reads to reduce time required to touch up
				if(counter==0){
					begReadIndex=new ReadIndex();
					begReadIndex.index=0;
					currReadIndex=begReadIndex;
				}else{
					currReadIndex.next=new ReadIndex();
					currReadIndex.next.prev=currReadIndex;
					currReadIndex=currReadIndex.next;
					currReadIndex.index=counter;
				}

				//identify the beginning and end positions of the read
				for(int j=0; j<nSites; j++){
					if(sequence.charAt(j)!='-'){ begPos[counter]=j; break; }
				}
				for(int j=(nSites-1); j>=0; j--){
					if(sequence.charAt(j)!='-'){ endPos[counter]=j; break; }
				}
				////////////////////////////////

				counter++;
			}		
			name=br.readLine();		
		}

		//NOW IDENTIFY THE CONSENSUS SEQUENCE (N IF <10 NON-N'S OR <90% OF THE NON-NS ARE IDENTICAL)
		//ALSO FIX N'S WHEN APPROPRIATE...
		char consensus[] = new char[nSites];
		int nA,nT,nC,nG;
		int nATCG=0;
		int nMax=0;
		int nErrors=0;
		boolean foundN=false;

		for(int j=0; j<nSites; j++){
			//count up the base frequences
			nA=nT=nC=nG=nATCG=nMax=0;		
			foundN=false;
			for(int i=0; i<nSeqs; i++){
				switch(alignment[i][j]){
					case 'A':	nA++; break;
					case 'T':	nT++; break;
					case 'C':	nC++; break;
					case 'G':	nG++; break;
					case 'N':	foundN=true;
				}
			}
			nATCG=nA+nT+nC+nG;
			//generate the consensus base
			if(nATCG<10){
				consensus[j]='N';
			}else{
				if(	 nA>=nATCG*0.9){	consensus[j]='A';
				}else if(nT>=nATCG*0.9){	consensus[j]='T';
				}else if(nC>=nATCG*0.9){	consensus[j]='C';	
				}else if(nG>=nATCG*0.9){	consensus[j]='G';
				}else{				consensus[j]='N';
				}
	
			}
	
			nMax=Math.max(Math.max(Math.max(nA,nT),nC),nG);
			nErrors=nATCG-nMax;
			
			//touchup if confident enough in the base call...
			for(int i=0; i<nSeqs; i++){
				if(alignment[i][j]=='N' && consensus[j]!='N'){		//touch up Ns if >=10 non N's and >=90% are the same base
					alignment[i][j]=consensus[j];	
				}else if(alignment[i][j]!='-' &&  nErrors <= maxErrorsAllowed[nATCG]){
					alignment[i][j]=consensus[j];
				}			
			}
			
		}	

		//NOW IDENTIFY THE SITES with  (TO BE REMOVED LATER)...
		boolean allGaps[] = new boolean[nSites];
		int nGaps=0;
		int nNonGaps=0;
		for(int j=0; j<nSites; j++){
			allGaps[j]=true;
			nGaps=nNonGaps=0;
			for(int i=0; i<nSeqs; i++){
				if(alignment[i][j]!='-'){
					if(j>0 && j>begPos[i] && j<endPos[i]){
						nNonGaps++;		//counts n gaps within reads as defined by having a base to the left or to the right
					}
					allGaps[j]=false;
				}else{
					if(j>0 && j>begPos[i] && j<endPos[i]){
						nGaps++;		//counts n gaps within reads as defined by having a base to the left or to the right
					}
				}
			}
			if(nGaps>0 && nGaps/(double)(nGaps+nNonGaps)>0.99){ allGaps[j]=true;}	//if site contains gap for >99% of overlapping reads
		}

		if(nSeqs>2){
				//ADJUST THE ORDER OF THE READS TO IMPROVE CLUMPYNESS FOR EASE OF MANUAL ADJUSTMENT
				currReadIndex = begReadIndex.next.next;
		
				ReadIndex tempReadIndex = currReadIndex;
				ReadIndex bestParent=null;
				int nMatches=0;
				int maxMatches=0;
				int prevMatches=0;
				while(currReadIndex!=null && currReadIndex.next!=null){
					tempReadIndex=currReadIndex;		
					currReadIndex=currReadIndex.next;	//ready for next move
					bestParent=tempReadIndex.prev;		//start with best parent is the current parent
					maxMatches=0;
					//identify the best paren
					tempReadIndex=bestParent;
					while(tempReadIndex!=null){		//move backwards up the list looking for a better match (parent), will break when overlap will no longer exist
						if(endPos[tempReadIndex.index]<=begPos[currReadIndex.prev.index]){break;}	//no sense checking if no overlap
						nMatches=0;
						for(int z=Math.max(begPos[tempReadIndex.index],begPos[currReadIndex.prev.index]); z<=Math.min(endPos[tempReadIndex.index],endPos[currReadIndex.prev.index]); z++){
							if(alignment[tempReadIndex.index][z]==alignment[currReadIndex.prev.index][z]){nMatches++;}
						}
						if(maxMatches==0){prevMatches=nMatches;}
						if(nMatches>maxMatches){maxMatches=nMatches; bestParent=tempReadIndex;}
						tempReadIndex=tempReadIndex.prev;
					}
					tempReadIndex=currReadIndex.prev;
		
					//move the read up
					if(bestParent!=tempReadIndex.prev){		//a better parent was found
		
						currReadIndex.prev=tempReadIndex.prev;
						currReadIndex.prev.next=currReadIndex;
						tempReadIndex.next=bestParent.next;
						bestParent.next.prev=tempReadIndex;
						bestParent.next=tempReadIndex;
						tempReadIndex.prev=bestParent;
					}			
		
				}
		}

		//write to a file
		int finalLocusLength=0;
		currReadIndex=begReadIndex;
		while(currReadIndex!=null){
			bw.write(seqHeaders[currReadIndex.index]+"\n");			
			for(int j=0; j<nSites; j++){
				if(!allGaps[j]){
					if(currReadIndex.index==0){finalLocusLength++;}
					bw.write(alignment[currReadIndex.index][j]);
				}
			}
			bw.write("\n");
			currReadIndex=currReadIndex.next;
		}
		bw.flush();
		bw.close();

		bw1.write(totalReads+"\n");
		bw2.write(nSeqs+"\n");
		bwL.write(finalLocusLength+"\n");
		bw1.flush();
		bw2.flush();
		bwL.flush();
	}
	System.out.println();
	bw1.flush();
	bw1.close();
	bw2.flush();
	bw2.close();
	bwL.flush();
	bwL.close();

      }catch(IOException ioe){System.out.println("<<!!ERROR main()!!>>");}
  }

  static void setBegPos(Pair pair){
  	for(int i=0; i<pair.read1.length(); i++){
  		if(pair.read1.charAt(i)!='-'){
  			pair.begPos=(i);		//note that this is 0 based...
  			return;
  		}
  	}
  }

  static void setEndPos(Pair pair){
	String read=pair.read1;		//default use end of first read as end of fragment
  	if(pair.read2!=null){		//if a second read was mapped use the end of hte second read as teh end of the fragment
  		read=pair.read2;
  	}
  	for(int i=read.length()-1; i>=0; i--){
  		if(read.charAt(i)!='-'){
  			pair.endPos=(i);				//note that this is 0 based...
			if(pair.read2!=null){
				pair.insertSize=pair.endPos-pair.begPos+1;
 			}
  			return;
  		}
  	}
  }
  
  static int getReadLength(String seq){
	if(seq==null){return -1;}
	int begPos=0;
  	for(int i=0; i<seq.length(); i++){
  		if(seq.charAt(i)!='-'){
  			begPos=i;
  			break;
  		}
  	}
	int endPos=0;
  	for(int i=seq.length()-1; i>=0; i--){
  		if(seq.charAt(i)!='-'){
  			endPos=i;
  			break;
  		}
  	}
	return endPos-begPos+1;
  }
///////////////////////////////////////////////////////
///////////////////////////////////////////////////////
  static double pbinomial(double q, int n, double p){
  	double result=0;
  	for(int x=0; x<q+1; x++){
  		result+=dbinomial(x,n,p,false);
  	}
  	return result;
  }
  static double dbinomial(int x, int n, double p, boolean returnLog){
  	double result = log_factorial(n) - (log_factorial(x)+log_factorial(n-x)) + (x * Math.log(p)) + ((n-x) * Math.log(1-p));
	if(returnLog){
		return result;
	}else{
		return Math.exp(result);
	}
  }
  static double log_factorial(int x){
  	//"Returns the logarithm of x!
  	return gammaln(x+1);
  }
  static double gammaln(int n){
    	//Logarithm of Euler's gamma function for discrete values. 	
  	if(n<1){
  		return Double.POSITIVE_INFINITY;
  	}else if(n<3){
  		return 0;
  	}else{
  		double c[]={76.18009172947146, -86.50532032941677, 24.01409824083091, -1.231739572450155, 0.001208650973866179, -0.5395239384953 * 0.00001};
  		double x = n;
  		double y = n;
  		double tm = x + 5.5;
  		tm -= (x + 0.5) * Math.log(tm);
  		double se =  1.0000000000000190015;
  		for(int j=0; j<6; j++){
  			y += 1.0;
  			se += c[j] / y;
  		}
  		return -tm + Math.log(2.5066282746310005 * se / x);
  	}
  
  
  }


}

class Pair{
	String name="";
	String library="";
	String read1=null;
	String read2=null;
	char side1='X';
	char side2='X';		//e.g. '1' or '2' corresponding to the read in the original pair or 'M' if merged...
	String MorE1="&";		//"Mapped" or "Extended"
	String MorE2="&";		//"Mapped" or "Extended"	
	int begPos=-1;		//position in the initial alignment that the predicted insert began
	int endPos=-1;			//position in the initial alignment that the predicted insert ended ...requires both reads
	int insertSize=-1;			//total predicted insert size...requires both reads
	Pair next=null;
	boolean chosen=false;
	Dup dup=null;
	boolean found=false;
	int timesObserved=0;
}

class Dup{
	int count=0;
	Pair bestPair=null;	//this is pointer to the pair that will be stored in the final alignment
}

class ReadIndex{
	int index=-1;
	ReadIndex next=null;
	ReadIndex prev=null;
}